package com.schbrain.canal.client.core;

import com.alibaba.otter.canal.client.CanalConnector;
import com.schbrain.canal.client.annotation.CanalEventListener;
import com.schbrain.canal.client.annotation.ListenPoint;
import com.schbrain.canal.client.annotation.TableFilter;
import com.schbrain.canal.client.conf.CanalClientConfig;
import com.schbrain.canal.client.conf.SchbrainCanalConfig;
import com.schbrain.canal.client.event.CanalEvent;
import com.schbrain.canal.client.transfer.TransponderFactory;
import com.schbrain.canal.client.utils.BeanUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.springframework.core.annotation.AnnotatedElementUtils;
import org.springframework.core.annotation.AnnotationUtils;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

@Slf4j
public class SimpleCanalClient extends AbstractCanalClient {
    /**
     * executor
     */
    private ThreadPoolExecutor executor;
    /**
     * 所有执行器
     */
    private final List<CanalEvent> listeners = new ArrayList<>();
    /**
     * 表过滤器
     */
    private Map<String,List<CanalEvent>> tableCanalEventMap = new HashMap<>();
    /**
     * 切面执行器
     */
    private final List<ListenerPoint> annoListeners = new ArrayList<>();

    public SimpleCanalClient(SchbrainCanalConfig canalConfig, TransponderFactory factory) {
        super(canalConfig,factory);
        executor = new ThreadPoolExecutor(0, Integer.MAX_VALUE, 60L, TimeUnit.SECONDS, new SynchronousQueue<>(), Executors.defaultThreadFactory());
        initListeners();
    }

    private void initListeners() {
        log.info("{}: initializing the listeners....", Thread.currentThread().getName());
        List<CanalEvent> list = BeanUtil.getBeansOfType(CanalEvent.class);
        if(list!=null && list.size() > 0){
            List<CanalEvent> unFilters = new ArrayList<>();
            for (CanalEvent canalEvent : list) {
                TableFilter filter = canalEvent.getClass().getAnnotation(TableFilter.class);
                if(filter!=null && StringUtils.isNotBlank(filter.schame()) && StringUtils.isNotBlank(filter.table())){
                    String key = filter.schame()+":"+filter.table();
                    List<CanalEvent> filterList = MapUtils.getObject(tableCanalEventMap,key,new ArrayList<>());
                    filterList.add(canalEvent);
                    continue;
                }
                unFilters.add(canalEvent);
            }
            if(unFilters!=null && unFilters.size()>0){
                listeners.addAll(unFilters);
            }
        }
        Map<String, Object> listenerMap = BeanUtil.getBeansWithAnnotation(CanalEventListener.class);
        if (listenerMap != null) {
            for (Object target : listenerMap.values()) {
                Method[] methods = target.getClass().getDeclaredMethods();
                if (methods != null && methods.length > 0) {
                    for (Method method : methods) {
                        ListenPoint l = AnnotatedElementUtils.findMergedAnnotation(method,ListenPoint.class);
                        if (l != null) {
                            annoListeners.add(new ListenerPoint(target, method, l));
                        }
                    }
                }
            }
        }
        log.info("{}: initializing the listeners end.", Thread.currentThread().getName());
        if (log.isWarnEnabled() && listeners.isEmpty() && annoListeners.isEmpty()) {
            log.warn("{}: No listener found in context! ", Thread.currentThread().getName());
        }
    }

    @Override
    protected void process(CanalConnector connector, Map.Entry<String, CanalClientConfig> config) {
        HandlerConf handlerConf = new HandlerConf(listeners,tableCanalEventMap,annoListeners);
        executor.submit(factory.newTransponder(connector, config,handlerConf));
    }

    @Override
    public void stop() {
        super.stop();
        executor.shutdown();
    }
}
